#include "dependency_tracking.h"

#include "runtime/assets.h"
#include "runtime/threading.h"

#include <stdbool.h>
#include <limits.h>
#include <assert.h>
#include <stdlib.h>
#include <string.h>


/* Track a list of dependencies per asset.
 * For the runtime, we only care about dependent -> dependency (so for example material -> texture),
 * but for assetc we also care about the opposite direction:
 * If an asset has the reprocess_on_dependency_change setting enabled,
 * we need to queue a reprocessing, even if the asset itself did not change.
 */

#define END_OF_LIST 0

/* 64 byte cache line - 8 (next index + count) / 4 (u32) = 14 */
#define BUCKET_ENTRY_COUNT 14
typedef struct vy_dep_list_bucket_s {
    uint32_t next;
    uint32_t count;
    vy_uid entries[BUCKET_ENTRY_COUNT];
} vy_dep_list_bucket;

typedef union {
    /* Indices of the first buckets */
    struct {
        uint32_t dependencies;
        uint32_t dependents;
    };
    uint32_t lists[2];
} vy_dep_list;

static vy_mutex *_guard;

static vy_dep_list_bucket *_buckets;
static uint32_t _bucket_count;
static uint32_t _bucket_capacity;


#define MAP_SIZE 2048
static vy_uid _uids[MAP_SIZE];
static vy_dep_list _lists[MAP_SIZE];

vy_result vyInitDependencyTracking(void) {
    _guard = vyCreateMutex();
    if (!_guard)
        return VY_UNKNOWN_ERROR;

    _buckets = malloc(sizeof(vy_dep_list_bucket) * 256);
    if (!_buckets)
        return VY_UNKNOWN_ERROR;
    _bucket_capacity = 256;
    _bucket_count    = 1;

    return VY_SUCCESS;
}

static uint32_t AllocNewBucket(void) {
    if (_bucket_count == _bucket_capacity) {
        void *t = realloc(_buckets, (size_t)_bucket_capacity * 2 * sizeof(vy_dep_list_bucket));
        if (!t)
            return 0;
        _buckets = t;
        _bucket_capacity *= 2;
    }
    memset(&_buckets[_bucket_count], 0, sizeof(vy_dep_list_bucket));
    return _bucket_count++;
}

static vy_result InsertIntoList(vy_uid list_asset, vy_uid uid, int list_index) {
    vyLockMutex(_guard);
    bool inserted = false;
    for (uint32_t i = 0; i < MAP_SIZE; i++) {
        uint32_t at = (list_asset + i) % MAP_SIZE;
        if (_uids[at] == list_asset || _uids[at] == 0) {
            _uids[at] = list_asset;

            /* Alloc a new list, if necessary */
            if (_lists[at].lists[list_index] == 0) {
                _lists[at].lists[list_index] = AllocNewBucket();
                if (!_lists[at].lists[list_index]) {
                    vyUnlockMutex(_guard);
                    return VY_UNKNOWN_ERROR;
                }
            }

            /* Advance to the end of the list */
            vy_dep_list_bucket *bucket = &_buckets[_lists[at].lists[list_index]];
            while (bucket->next != END_OF_LIST) {
                bucket = &_buckets[bucket->next];
            }

            /* Grow the list, if necessary */
            if (bucket->count == BUCKET_ENTRY_COUNT) {
                bucket->next = AllocNewBucket();
                if (!bucket->next) {
                    vyUnlockMutex(_guard);
                    return VY_UNKNOWN_ERROR;
                }
                bucket = &_buckets[bucket->next];
            }
            
            assert(bucket->count < BUCKET_ENTRY_COUNT);
            bucket->entries[bucket->count++] = uid;

            inserted = true;
            break;
        }
    }
    vyUnlockMutex(_guard);
    assert(inserted);
    return VY_SUCCESS;
}

vy_result vyAddAssetDependency(vy_uid dependent, vy_uid dependency) {
    vy_result res = InsertIntoList(dependent, dependency, 0);
    if (res != VY_SUCCESS)
        return res;
    res = InsertIntoList(dependency, dependent, 1);
    return res;
}
